-module(gmc_v_devman).
-vsn("0.2.0").
-author("Craig Everett <craigeverett@qpq.swiss>").
-copyright("QPQ AG <info@qpq.swiss>").
-license("GPL-3.0-or-later").

-behavior(wx_object).
%-behavior(gmc_v).
-include_lib("wx/include/wx.hrl").
-export([to_front/1]).
-export([set_manifest/1, trouble/1]).
-export([start_link/1]).
-export([init/1, terminate/2, code_change/3,
         handle_call/3, handle_cast/2, handle_info/2, handle_event/2]).
-include("$zx_include/zx_logger.hrl").
-include("gmc.hrl").

% Widgets
-record(w,
        {name = none :: atom() | {FunName :: binary(), call | dryr},
         id   = 0    :: integer(),
         wx   = none :: none | wx:wx_object()}).

% Contract functions in an ACI
-record(f,
        {name   = <<"">> :: binary(),
         call   = #w{}   :: #w{},
         dryrun = #w{}   :: none | #w{},
         args   = []     :: [{wx:wx_object(), wx:wx_object(), argt()}]}).

% Code book pages
-record(p,
        {path      = {file, ""} :: {file, file:filename()} | {hash, binary()},
         win       = none       :: none | wx:wx_object(),
         code      = none       :: none | wxStyledTextCtrl:wxStyledTextCtrl()}).

% Contract pages
-record(c,
        {id        = ""         :: string(),
         win       = none       :: none | wx:wx_object(),
         code      = none       :: none | wxStyledTextCtrl:wxStyledTextCtrl(),
         cons      = none       :: none | wxStyledTextCtrl:wxStyledTextCtrl(),
         funs      = {#w{}, []} :: {#w{}, [#f{}]}}).

% State
-record(s,
        {wx      = none       :: none | wx:wx_object(),
         frame   = none       :: none | wx:wx_object(),
         lang    = en         :: en | jp,
         j       = none       :: none | fun(),
         prefs   = #{}        :: map(),
         buttons = #{}        :: #{WX_ID :: integer() := #w{}},
         tabs    = none       :: none | wx:wx_object(),
         code    = {none, []} :: {Codebook :: none | wx:wx_object(), Pages :: [#p{}]},
         cons    = {none, []} :: {Consbook :: none | wx:wx_object(), Pages :: [#c{}]}}).

-type argt() :: int | string | address | list(argt()).

%%% Interface

-spec to_front(Win) -> ok
    when Win :: wx:wx_object().

to_front(Win) ->
    wx_object:cast(Win, to_front).


-spec set_manifest(Entries) -> ok
    when Entries :: [ael:conf_meta()].

set_manifest(Entries) ->
    case is_pid(whereis(?MODULE)) of
        true  -> wx_object:cast(?MODULE, {set_manifest, Entries});
        false -> ok
    end.


-spec trouble(Info) -> ok
    when Info :: term().

trouble(Info) ->
    wx_object:cast(?MODULE, {trouble, Info}).




%%% Startup Functions

start_link(Args) ->
    wx_object:start_link({local, ?MODULE}, ?MODULE, Args, []).


init({Prefs, Manifest}) ->
    Lang = maps:get(lang, Prefs, en_us),
    Trans = gmc_jt:read_translations(?MODULE),
    J = gmc_jt:j(Lang, Trans),
    Wx = wx:new(),
    Frame = wxFrame:new(Wx, ?wxID_ANY, J("Contracts")),

    MainSz = wxBoxSizer:new(?wxVERTICAL),
    TopBook = wxNotebook:new(Frame, ?wxID_ANY, [{style, ?wxBK_DEFAULT}]),
    {LWin, LButtons, Codebook} = make_l_win(TopBook, J),
    {RWin, RButtons, Consbook} = make_r_win(TopBook, J),
    ButtonMap = maps:merge(LButtons, RButtons),
    true = wxNotebook:addPage(TopBook, LWin, J("Contract Editor"), []),
    true = wxNotebook:addPage(TopBook, RWin, J("Deployed Contracts"), []),
    State =
        #s{wx      = Wx,             frame  = Frame,
           j       = J,              prefs  = Prefs,
           buttons = ButtonMap,      tabs   = TopBook,
           code    = {Codebook, []}, cons   = {Consbook, []}},
    _ = wxSizer:add(MainSz, TopBook, zxw:flags(wide)),
    _ = wxFrame:setSizer(Frame, MainSz),
    _ = wxSizer:layout(MainSz),
    ok = gmc_v:safe_size(Frame, Prefs),
    ok = wxFrame:connect(Frame, close_window),
    ok = wxFrame:connect(Frame, command_button_clicked),
    ok = wxFrame:center(Frame),
    true = wxFrame:show(Frame),
    NewState = add_code_pages(State, Manifest),
    {Frame, NewState}.


make_l_win(TopBook, J) ->
    Win = wxWindow:new(TopBook, ?wxID_ANY),
    MainSz = wxBoxSizer:new(?wxVERTICAL),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    ButtonTemplates =
        [{new,          J("New")},
         {open,         J("Open")},
         {save,         J("Save")},
         {rename,       J("Save (rename)")},
         {deploy,       J("Deploy")},
         {close_source, J("Close")}],
    MakeButton =
        fun({Name, Label}) ->
            B = wxButton:new(Win, ?wxID_ANY, [{label, Label}]),
            #w{name = Name, id = wxButton:getId(B), wx = B}
        end,
    Buttons = lists:map(MakeButton, ButtonTemplates),
    AddButton = fun(#w{wx = B}) -> wxSizer:add(ButtSz, B, zxw:flags(wide)) end,
    Codebook = wxNotebook:new(Win, ?wxID_ANY, [{style, ?wxBK_DEFAULT}]),
    ok = lists:foreach(AddButton, Buttons),
    _ = wxSizer:add(MainSz, ButtSz,   zxw:flags(base)),
    _ = wxSizer:add(MainSz, Codebook, zxw:flags(wide)),
    _ = wxWindow:setSizer(Win, MainSz),
    MapButton = fun(B = #w{id = I}, M) -> maps:put(I, B, M) end,
    ButtonMap = lists:foldl(MapButton, #{}, Buttons),
    {Win, ButtonMap, Codebook}.

make_r_win(TopBook, J) ->
    Win = wxWindow:new(TopBook, ?wxID_ANY),
    MainSz = wxBoxSizer:new(?wxVERTICAL),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    ButtonTemplates =
        [{load,           J("Load from Chain")},
         {edit,           J("Copy to Editor")},
         {close_instance, J("Close")}],
    MakeButton =
        fun({Name, Label}) ->
            B = wxButton:new(Win, ?wxID_ANY, [{label, Label}]),
            #w{name = Name, id = wxButton:getId(B), wx = B}
        end,
    Buttons = lists:map(MakeButton, ButtonTemplates),
    AddButton = fun(#w{wx = B}) -> wxSizer:add(ButtSz, B, zxw:flags(wide)) end,
    Codebook = wxNotebook:new(Win, ?wxID_ANY, [{style, ?wxBK_DEFAULT}]),
    ok = lists:foreach(AddButton, Buttons),
    _ = wxSizer:add(MainSz, ButtSz,   zxw:flags(base)),
    _ = wxSizer:add(MainSz, Codebook, zxw:flags(wide)),
    _ = wxWindow:setSizer(Win, MainSz),
    MapButton = fun(B = #w{id = I}, M) -> maps:put(I, B, M) end,
    ButtonMap = lists:foldl(MapButton, #{}, Buttons),
    {Win, ButtonMap, Codebook}.



add_code_pages(State, Files) ->
    lists:foldl(fun add_code_page/2, State, Files).



%%% OTP callbacks

handle_call(Unexpected, From, State) ->
    ok = log(warning, "Unexpected call from ~tp: ~tp~n", [From, Unexpected]),
    {noreply, State}.


handle_cast(to_front, State = #s{frame = Frame}) ->
    ok = wxFrame:raise(Frame),
    {noreply, State};
handle_cast({trouble, Info}, State) ->
    ok = handle_troubling(State, Info),
    {noreply, State};
handle_cast(Unexpected, State) ->
    ok = log(warning, "Unexpected cast: ~tp~n", [Unexpected]),
    {noreply, State}.


handle_info(Unexpected, State) ->
    ok = log(warning, "Unexpected info: ~tp~n", [Unexpected]),
    {noreply, State}.


handle_event(E = #wx{event = #wxCommand{type = command_button_clicked},
                     id    = ID},
             State = #s{buttons = Buttons}) ->
    NewState =
        case maps:get(ID, Buttons, undefined) of
            #w{name = new}               -> new_file(State);
            #w{name = open}              -> open(State);
            #w{name = save}              -> save(State);
            #w{name = rename}            -> rename(State);
            #w{name = deploy}            -> deploy(State);
            #w{name = close_source}      -> close_source(State);
            #w{name = load}              -> load(State);
            #w{name = edit}              -> edit(State);
            #w{name = close_instance}    -> close_instance(State);
            #w{name = Name, wx = Button} -> clicked(State, Name, Button);
            undefined ->
                tell("Received message: ~w", [E]),
                State
        end,
    {noreply, NewState};
handle_event(#wx{event = #wxClose{}}, State = #s{frame = Frame, prefs = Prefs}) ->
    Geometry =
        case wxTopLevelWindow:isMaximized(Frame) of
            true ->
                max;
            false ->
                {X, Y} = wxWindow:getPosition(Frame),
                {W, H} = wxWindow:getSize(Frame),
                {X, Y, W, H}
        end,
    NewPrefs = maps:put(geometry, Geometry, Prefs),
    ok = gmc_con:save(?MODULE, NewPrefs),
    ok = wxWindow:destroy(Frame),
    {noreply, State};
handle_event(Event, State) ->
    ok = tell(info, "Unexpected event ~tp State: ~tp~n", [Event, State]),
    {noreply, State}.


handle_troubling(#s{frame = Frame}, Info) ->
    zxw:show_message(Frame, Info).


code_change(_, State, _) ->
    {ok, State}.


terminate(Reason, State) ->
    ok = log(info, "Reason: ~tp, State: ~tp", [Reason, State]),
    wx:destroy().



%%% Doers

clicked(State = #s{cons = {Consbook, Contracts}}, Name, Button) ->
    case wxNotebook:getSelection(Consbook) of
        ?wxNOT_FOUND ->
            ok = tell(warning, "Inconcievable! No notebook page is selected!"),
            State;
        Index ->
            Contract = lists:nth(Index + 1, Contracts),
            clicked(State, Contract, Name, Button)
    end.

clicked(State, Contract, Name, Button) ->
    ok = tell("Button: ~p ~p~nContract: ~p", [Button, Name, Contract]),
    State.

get_arg({_, TextCtrl, _}) ->
    wxTextCtrl:getValue(TextCtrl).


add_code_page(State = #s{code = {Codebook, Pages}}, File) ->
    case keyfind_index({file, File}, #p.path, Pages) of
        error ->
            add_code_page2(State, File);
        {ok, Index} ->
            _ = wxNotebook:setSelection(Codebook, Index - 1),
            State
    end.

add_code_page2(State = #s{j = J}, {file, File}) ->
    case file:read_file(File) of
        {ok, Bin} ->
            case unicode:characters_to_list(Bin) of
                Code when is_list(Code) ->
                    add_code_page(State, {file, File}, Code);
                Error ->
                    Message = io_lib:format(J("Opening ~p failed with: ~p"), [File, Error]),
                    ok = handle_troubling(Message, State),
                    State
            end;
        {error, Reason} ->
            Message = io_lib:format(J("Opening ~p failed with: ~p"), [File, Reason]),
            ok = handle_troubling(Message, State),
            State
    end;
add_code_page2(State, {hash, Address}) ->
    open_hash2(State, Address).

add_code_page(State = #s{j = J, tabs = TopBook, code = {Codebook, Pages}}, Location, Code) ->
%   FIXME: One of these days we need to define the text area as a wxStyledTextCtrl and will
%          have to contend with system theme issues (light/dark themese, namely)
%          Leaving this little thing here to remind myself how any of that works later.
%          The call below returns a wx_color4() type (not that we need alpha...).
%   Color = wxSystemSettings:getColour(?wxSYS_COLOUR_WINDOW),
%   tell("Color: ~p", [Color]),
    Window = wxWindow:new(Codebook, ?wxID_ANY),
    PageSz = wxBoxSizer:new(?wxHORIZONTAL),

    CodeTxStyle = {style, ?wxTE_MULTILINE bor ?wxTE_PROCESS_TAB bor ?wxTE_DONTWRAP},
    CodeTx = wxTextCtrl:new(Window, ?wxID_ANY, [CodeTxStyle]),
    TextAt = wxTextAttr:new(),
    Mono = wxFont:new(10, ?wxMODERN, ?wxNORMAL, ?wxNORMAL, [{face, "Monospace"}]),
    ok = wxTextAttr:setFont(TextAt, Mono),
    true = wxTextCtrl:setDefaultStyle(CodeTx, TextAt),
    ok = wxTextCtrl:setValue(CodeTx, Code),

    _ = wxSizer:add(PageSz, CodeTx, zxw:flags(wide)),

    ok = wxWindow:setSizer(Window, PageSz),
    ok = wxSizer:layout(PageSz),
    _ = wxNotebook:changeSelection(TopBook, 0),
    FileName =
        case Location of
            {file, Path} -> filename:basename(Path);
            {hash, Addr} -> Addr
        end,
    true = wxNotebook:addPage(Codebook, Window, FileName, [{bSelect, true}]),
    Page = #p{path = Location, win = Window, code = CodeTx},
    NewPages = Pages ++ [Page],
    State#s{code = {Codebook, NewPages}}.


new_file(State = #s{frame = Frame, j = J, prefs = Prefs}) ->
    DefaultDir =
        case maps:find(dir, Prefs) of
            {ok, PrefDir} ->
                PrefDir;
            error ->
                case os:getenv("ZOMP_DIR") of
                    "" -> file:get_pwd();
                    D  -> filename:basename(D)
                end
        end,
    Options =
        [{message,     J("Save Location")},
         {defaultDir,  DefaultDir},
         {defaultFile, "my_contract.aes"},
         {wildCard,    "*.aes"},
         {style,       ?wxFD_SAVE bor ?wxFD_OVERWRITE_PROMPT}],
    Dialog = wxFileDialog:new(Frame, Options),
    NewState =
        case wxFileDialog:showModal(Dialog) of
            ?wxID_OK ->
                Dir = wxFileDialog:getDirectory(Dialog),
                case wxFileDialog:getFilename(Dialog) of
                    "" ->
                        State;
                    Name ->
                        File =
                            case filename:extension(Name) of
                                ".aes" -> Name;
                                _      -> Name ++ ".aes"
                            end,
                        Path = filename:join(Dir, File),
                        NewPrefs = maps:put(dir, Dir, Prefs),
                        NextState = State#s{prefs = NewPrefs},
                        add_code_page(NextState, {file, Path}, "")
                end;
            ?wxID_CANCEL ->
                State
        end,
    ok = wxFileDialog:destroy(Dialog),
    NewState.


deploy(State = #s{code = {Codebook, Pages}}) ->
    case wxNotebook:getSelection(Codebook) of
        ?wxNOT_FOUND ->
            State;
        Index ->
            Page = #p{code = CodeTx} = lists:nth(Index + 1, Pages),
            Source = wxTextCtrl:getValue(CodeTx),
            deploy2(State, Source)
    end.

deploy2(State, Source) ->
    case aeso_compiler:from_string(Source, [{aci, json}]) of
        {ok, Build = #{aci := ACI}} ->
            FunDefs = {#{functions := Funs}, _} = find_main(ACI),
            Init = lom:find(name, <<"init">>, Funs),
            ok = tell(info, "Compilation Succeeded!~n~tp~n~n~tp", [Build, FunDefs]),
            deploy3(State, Init);
        Other ->
            ok = tell(info, "Compilation Failed!~n~tp", [Other]),
            State
    end.

deploy3(State = #s{frame = Frame, j = J}, #{arguments := As}) ->
    Dialog = wxDialog:new(Frame, ?wxID_ANY, J("Deploy Contract")),
    Sizer = wxBoxSizer:new(?wxVERTICAL),
    ScrollWin = wxScrolledWindow:new(Dialog),
    FunName = unicode:characters_to_list(["init/", integer_to_list(length(As))]), 
    FunSizer = wxStaticBoxSizer:new(?wxVERTICAL, ScrollWin, [{label, FunName}]),
    ok = wxScrolledWindow:setSizerAndFit(ScrollWin, FunSizer),
    ok = wxScrolledWindow:setScrollRate(ScrollWin, 5, 5),
    ButtSz = wxDialog:createButtonSizer(Dialog, ?wxOK bor ?wxCANCEL),
    GridSz = wxFlexGridSizer:new(2, 4, 4),
    ok = wxFlexGridSizer:setFlexibleDirection(GridSz, ?wxHORIZONTAL),
    ok = wxFlexGridSizer:addGrowableCol(GridSz, 1),
    MakeArgField =
        fun(#{name := AN, type := T}) ->
                Type =
                    case T of
                        <<"address">>       -> address;
                        <<"int">>           -> integer;
                        <<"bool">>          -> boolean;
                        L when is_list(L)   -> list;   % FIXME
%                       I when is_binary(I) -> iface   % FIXME
                        I when is_binary(I) -> address % FIXME
                    end,
                ANT = wxStaticText:new(ScrollWin, ?wxID_ANY, AN),
                TCT = wxTextCtrl:new(ScrollWin, ?wxID_ANY),
                _ = wxFlexGridSizer:add(GridSz, ANT, zxw:flags(base)),  
                _ = wxFlexGridSizer:add(GridSz, TCT, zxw:flags(wide)),  
                {ANT, TCT, Type}
        end,
    ArgFields = lists:map(MakeArgField, As),
    _ = wxStaticBoxSizer:add(FunSizer, GridSz, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, ScrollWin, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, ButtSz),
    ok = wxDialog:setSizer(Dialog, Sizer),
    ok = wxBoxSizer:layout(Sizer),
    ok = wxDialog:setSize(Dialog, {500, 300}),
    ok = wxDialog:center(Dialog),
    ok =
        case wxDialog:showModal(Dialog) of
            ?wxID_OK ->
                tell(info, "DEPLOYING!");
            ?wxID_CANCEL ->
                ok
        end,
    ok = wxDialog:destroy(Dialog),
    State.

find_main(ACI) ->
    find_main(ACI, none, []).

find_main([#{contract := I = #{kind := contract_interface}} | T], M, Is) ->
    find_main(T, M, [I | Is]);
find_main([#{contract := M = #{kind := contract_main}}      | T], _, Is) ->
    find_main(T, M, Is);
find_main([#{namespace := _} | T], M, Is) ->
    find_main(T, M, Is);
find_main([C | T], M, Is) ->
    ok = tell("Surprising ACI element: ~p", [C]),
    find_main(T, M, Is);
find_main([], M, Is) ->
    {M, Is}.

fun_interfaces(Window,
               Buttons,
               {OldScrollWin, OldIfaces},
               {#{name := Name, functions := Funs}, _ConIfaces},
               J) ->
    ok = wxScrolledWindow:destroy(OldScrollWin),
    OldButtonIDs = button_key_list(OldIfaces),
    NextButtons = maps:without(OldButtonIDs, Buttons),
    ScrollWin = wxScrolledWindow:new(Window),
    FSOpts = [{label, J("Function Interfaces")}],
    FunSizer = wxStaticBoxSizer:new(?wxVERTICAL, ScrollWin, FSOpts),
    ok = wxScrolledWindow:setSizerAndFit(ScrollWin, FunSizer),
    ok = wxScrolledWindow:setScrollRate(ScrollWin, 5, 5),
    ConName = wxStaticText:new(ScrollWin, ?wxID_ANY, Name),
    _ = wxSizer:add(FunSizer, ConName),
    MakeIface =
        fun(#{name := N, arguments := As}) ->
            FunName = unicode:characters_to_list([N, "/", integer_to_list(length(As))]), 
            FN = wxStaticBoxSizer:new(?wxVERTICAL, ScrollWin, [{label, FunName}]),
            GridSz = wxFlexGridSizer:new(2, 4, 4),
            ok = wxFlexGridSizer:setFlexibleDirection(GridSz, ?wxHORIZONTAL),
            ok = wxFlexGridSizer:addGrowableCol(GridSz, 1),
            MakeArgField =
                fun(#{name := AN, type := T}) ->
                        Type =
                            case T of
                                <<"address">>       -> address;
                                <<"int">>           -> integer;
                                <<"bool">>          -> boolean;
                                L when is_list(L)   -> list;   % FIXME
%                               I when is_binary(I) -> iface   % FIXME
                                I when is_binary(I) -> address % FIXME
                            end,
                        ANT = wxStaticText:new(ScrollWin, ?wxID_ANY, AN),
                        TCT = wxTextCtrl:new(ScrollWin, ?wxID_ANY),
                        _ = wxFlexGridSizer:add(GridSz, ANT, zxw:flags(base)),  
                        _ = wxFlexGridSizer:add(GridSz, TCT, zxw:flags(wide)),  
                        {ANT, TCT, Type}
                end,
            ArgFields = lists:map(MakeArgField, As),
            ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
            {CallButton, DryRunButton} =
                case N =:= <<"init">> of
                    false ->
                        CallBn = wxButton:new(ScrollWin, ?wxID_ANY, [{label, J("Call")}]),
                        DryRBn = wxButton:new(ScrollWin, ?wxID_ANY, [{label, J("Dry Run")}]),
                        true = wxButton:disable(CallBn),
                        true = wxButton:disable(DryRBn),
                        _ = wxBoxSizer:add(ButtSz, CallBn, zxw:flags(wide)),
                        _ = wxBoxSizer:add(ButtSz, DryRBn, zxw:flags(wide)),
                        {#w{name = {N, call}, id = wxButton:getId(CallBn), wx = CallBn},
                         #w{name = {N, dryr}, id = wxButton:getId(DryRBn), wx = DryRBn}};
                    true ->
                        Deploy = wxButton:new(ScrollWin, ?wxID_ANY, [{label, J("Deploy")}]),
                        _ = wxBoxSizer:add(ButtSz, Deploy, zxw:flags(wide)),
                        {#w{name = {N, call}, id = wxButton:getId(Deploy), wx = Deploy},
                         none}
                end,
            _ = wxStaticBoxSizer:add(FN, GridSz, zxw:flags(wide)),
            _ = wxStaticBoxSizer:add(FN, ButtSz, zxw:flags(base)),
            _ = wxSizer:add(FunSizer, FN, zxw:flags(base)),
            #f{name = N, call = CallButton, dryrun = DryRunButton, args = ArgFields}
        end,
    Ifaces = lists:map(MakeIface, Funs),
    NewButtons = lists:foldl(fun map_iface_buttons/2, NextButtons, Ifaces),
    ok = wxSizer:layout(FunSizer),
    {NewButtons, {ScrollWin, Ifaces}}.

button_key_list([#f{call = #w{id = C}, dryrun = #w{id = D}} | T]) ->
    [C, D | button_key_list(T)];
button_key_list([#f{call = #w{id = C}, dryrun = none} | T]) ->
    [C | button_key_list(T)];
button_key_list([]) ->
    [].

map_iface_buttons(#f{call = C = #w{id = CID}, dryrun = D = #w{id = DID}}, A) ->
    maps:put(DID, D, maps:put(CID, C, A));
map_iface_buttons(#f{call = C = #w{id = CID}, dryrun = none}, A) ->
    maps:put(CID, C, A).


open(State = #s{frame = Frame, j = J}) ->
    Dialog = wxDialog:new(Frame, ?wxID_ANY, J("Open Contract Source")),
    Sizer = wxBoxSizer:new(?wxVERTICAL),
    Choices = wxRadioBox:new(Dialog,
                             ?wxID_ANY,
                             J("Select Origin"),
                             ?wxDefaultPosition,
                             ?wxDefaultSize,
                             [J("From File"), J("From Hash")],
                             [{majorDim, 1}]),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    Affirm = wxButton:new(Dialog, ?wxID_OK),
    Cancel = wxButton:new(Dialog, ?wxID_CANCEL),
    _ = wxBoxSizer:add(ButtSz, Affirm, zxw:flags(wide)),
    _ = wxBoxSizer:add(ButtSz, Cancel, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, Choices, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, ButtSz, zxw:flags(wide)),
    ok = wxDialog:setSizer(Dialog, Sizer),
    ok = wxDialog:setSize(Dialog, {250, 170}),
    ok = wxBoxSizer:layout(Sizer),
    Choice =
        case wxDialog:showModal(Dialog) of
            ?wxID_OK ->
                case wxRadioBox:getSelection(Choices) of
                    0            -> file;
                    1            -> hash;
                    ?wxNOT_FOUND -> none
                end;
            ?wxID_CANCEL -> cancel
        end,
    ok = wxDialog:destroy(Dialog),
    case Choice of
        file ->
            open_file(State);
        hash ->
            open_hash(State);
        none ->
            ok = tell(info, "No selection."),
            State;
        cancel ->
            ok = tell(info, "Cancelled."),
            State
    end.


open_file(State = #s{frame = Frame, j = J, prefs = Prefs}) ->
    DefaultDir =
        case maps:find(dir, Prefs) of
            {ok, PrefDir} ->
                PrefDir;
            error ->
                case os:getenv("ZOMP_DIR") of
                    "" -> file:get_pwd();
                    D  -> filename:basename(D)
                end
        end,
    Options =
        [{message,     J("Load Contract Source")},
         {defaultDir,  DefaultDir},
         {wildCard,    "*.aes"},
         {style,       ?wxFD_OPEN}],
    Dialog = wxFileDialog:new(Frame, Options),
    NewState =
        case wxFileDialog:showModal(Dialog) of
            ?wxID_OK ->
                Dir = wxFileDialog:getDirectory(Dialog),
                case wxFileDialog:getFilename(Dialog) of
                    "" ->
                        State;
                    Name ->
                        File =
                            case filename:extension(Name) of
                                ".aes" -> Name;
                                _      -> Name ++ ".aes"
                            end,
                        Path = filename:join(Dir, File),
                        NewPrefs = maps:put(dir, Dir, Prefs),
                        NextState = State#s{prefs = NewPrefs},
                        add_code_page(NextState, {file, Path})
                end;
            ?wxID_CANCEL ->
                State
        end,
    ok = wxFileDialog:destroy(Dialog),
    NewState.


open_hash(State = #s{frame = Frame, j = J}) ->
    Dialog = wxDialog:new(Frame, ?wxID_ANY, J("Retrieve Contract Source")),
    Sizer = wxBoxSizer:new(?wxVERTICAL),
    AddressSz = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Address Hash")}]),
    AddressTx = wxTextCtrl:new(Dialog, ?wxID_ANY),
    _ = wxSizer:add(AddressSz, AddressTx, zxw:flags(wide)),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    Affirm = wxButton:new(Dialog, ?wxID_OK),
    Cancel = wxButton:new(Dialog, ?wxID_CANCEL),
    _ = wxBoxSizer:add(ButtSz, Affirm, zxw:flags(wide)),
    _ = wxBoxSizer:add(ButtSz, Cancel, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, AddressSz, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, ButtSz, zxw:flags(wide)),
    ok = wxDialog:setSizer(Dialog, Sizer),
    ok = wxBoxSizer:layout(Sizer),
    ok = wxDialog:setSize(Dialog, {500, 200}),
    ok = wxDialog:center(Dialog),
    ok = wxStyledTextCtrl:setFocus(AddressTx),
    Choice =
        case wxDialog:showModal(Dialog) of
            ?wxID_OK ->
                case wxTextCtrl:getValue(AddressTx) of
                    "" -> cancel;
                    A  -> {ok, A}
                end;
            ?wxID_CANCEL ->
                cancel
        end,
    ok = wxDialog:destroy(Dialog),
    case Choice of
        {ok, Address} -> open_hash2(State, Address);
        cancel        -> State
    end.

open_hash2(State, Address) ->
    case hz:contract_source(Address) of
        {ok, Source} ->
            open_hash3(State, Address, Source);
        Error ->
            ok = handle_troubling(Error, State),
            State
    end.

open_hash3(State, Address, Source) ->
    % TODO: Compile on load and verify the deployed hash for validity.
    case aeso_compiler:from_string(Source, [{aci, json}]) of
        {ok, Build = #{aci := ACI}} ->
            {Defs = #{functions := Funs}, ConIfaces} = find_main(ACI),
            Callable = lom:delete(name, <<"init">>, Funs),
            FunDefs = {maps:put(functions, Callable, Defs), ConIfaces},
            ok = tell(info, "Compilation Succeeded!~n~tp~n~n~tp", [Build, FunDefs]),
            add_code_page(State, {hash, Address}, Source);
        Other ->
            ok = tell(info, "Compilation Failed!~n~tp", [Other]),
            State
    end.


% TODO: Break this down -- tons of things in here recur.
save(State = #s{frame = Frame, j = J, prefs = Prefs, code = {Codebook, Pages}}) ->
    case wxNotebook:getSelection(Codebook) of
        ?wxNOT_FOUND ->
            State;
        Index ->
            case lists:nth(Index + 1, Pages) of
                #p{path = {file, Path}, code = Widget} ->
                    Source = wxStyledTextCtrl:getText(Widget),
                    case filelib:ensure_dir(Path) of
                        ok ->
                            case file:write_file(Path, Source) of
                                ok ->
                                    State;
                                Error ->
                                    ok = handle_troubling(State, Error),
                                    State
                            end;
                        Error ->
                            ok = handle_troubling(State, Error),
                            State
                    end;
                Page = #p{path = {hash, Hash}, code = Widget} ->
                    DefaultDir =
                        case maps:find(dir, Prefs) of
                            {ok, PrefDir} ->
                                PrefDir;
                            error ->
                                case os:getenv("ZOMP_DIR") of
                                    "" -> file:get_pwd();
                                    D  -> filename:basename(D)
                                end
                        end,
                    Options =
                        [{message,     J("Save Location")},
                         {defaultDir,  DefaultDir},
                         {defaultFile, unicode:characters_to_list([Hash, ".aes"])},
                         {wildCard,    "*.aes"},
                         {style,       ?wxFD_SAVE bor ?wxFD_OVERWRITE_PROMPT}],
                    Dialog = wxFileDialog:new(Frame, Options),
                    NewState =
                        case wxFileDialog:showModal(Dialog) of
                            ?wxID_OK ->
                                Dir = wxFileDialog:getDirectory(Dialog),
                                case wxFileDialog:getFilename(Dialog) of
                                    "" ->
                                        State;
                                    Name ->
                                        File =
                                            case filename:extension(Name) of
                                                ".aes" -> Name;
                                                _      -> Name ++ ".aes"
                                            end,
                                        Path = filename:join(Dir, File),
                                        Source = wxTextCtrl:getValue(Widget),
                                        case filelib:ensure_dir(Path) of
                                            ok ->
                                                case file:write_file(Path, Source) of
                                                    ok ->
                                                        true = wxNotebook:setPageText(Codebook, Index, File),
                                                        NewPrefs = maps:put(dir, Dir, Prefs),
                                                        NewPage  = Page#p{path = {file, Path}},
                                                        NewPages = store_nth(Index + 1, NewPage, Pages),
                                                        NewCode = {Codebook, NewPages},
                                                        State#s{prefs = NewPrefs, code = NewCode};
                                                    Error ->
                                                        ok = handle_troubling(State, Error),
                                                        State
                                                end;
                                            Error ->
                                                ok = handle_troubling(State, Error),
                                                State
                                        end
                                end;
                            ?wxID_CANCEL ->
                                State
                        end,
                    ok = wxFileDialog:destroy(Dialog),
                    NewState
            end
    end.

% TODO: Break this down -- tons of things in here recur.
rename(State = #s{frame = Frame, j = J, prefs = Prefs, code = {Codebook, Pages}}) ->
    case wxNotebook:getSelection(Codebook) of
        ?wxNOT_FOUND ->
            State;
        Index ->
            case lists:nth(Index + 1, Pages) of
                Page = #p{path = {file, Path}, code = Widget} ->
                    DefaultDir = filename:dirname(Path),
                    Options =
                        [{message,     J("Save Location")},
                         {defaultDir,  DefaultDir},
                         {defaultFile, filename:basename(Path)},
                         {wildCard,    "*.aes"},
                         {style,       ?wxFD_SAVE bor ?wxFD_OVERWRITE_PROMPT}],
                    Dialog = wxFileDialog:new(Frame, Options),
                    NewState =
                        case wxFileDialog:showModal(Dialog) of
                            ?wxID_OK ->
                                Dir = wxFileDialog:getDirectory(Dialog),
                                case wxFileDialog:getFilename(Dialog) of
                                    "" ->
                                        State;
                                    Name ->
                                        File =
                                            case filename:extension(Name) of
                                                ".aes" -> Name;
                                                _      -> Name ++ ".aes"
                                            end,
                                        NewPath = filename:join(Dir, File),
                                        Source = wxTextCtrl:getValue(Widget),
                                        case filelib:ensure_dir(NewPath) of
                                            ok ->
                                                case file:write_file(NewPath, Source) of
                                                    ok ->
                                                        true = wxNotebook:setPageText(Codebook, Index, File),
                                                        NewPrefs = maps:put(dir, Dir, Prefs),
                                                        NewPage  = Page#p{path = {file, NewPath}},
                                                        NewPages = store_nth(Index + 1, NewPage, Pages),
                                                        NewCode = {Codebook, NewPages},
                                                        State#s{prefs = NewPrefs, code = NewCode};
                                                    Error ->
                                                        ok = handle_troubling(State, Error),
                                                        State
                                                end;
                                            Error ->
                                                ok = handle_troubling(State, Error),
                                                State
                                        end
                                end;
                            ?wxID_CANCEL ->
                                State
                        end,
                    ok = wxFileDialog:destroy(Dialog),
                    NewState;
                #p{path = {hash, _}} ->
                    save(State)
            end
    end.


close_source(State = #s{code = {Codebook, Pages}}) ->
    case wxNotebook:getSelection(Codebook) of
        ?wxNOT_FOUND ->
            State;
        Index ->
            NewPages = drop_nth(Index + 1, Pages),
            true = wxNotebook:deletePage(Codebook, Index),
            State#s{code = {Codebook, NewPages}}
    end.


load(State = #s{frame = Frame, j = J}) ->
    % TODO: Extract the exact compiler version, load it, and use only that or fail if
    %       the specific version is unavailable.
    % TODO: Compile on load and verify the deployed hash for validity.
    Dialog = wxDialog:new(Frame, ?wxID_ANY, J("Retrieve Contract Source")),
    Sizer = wxBoxSizer:new(?wxVERTICAL),
    AddressSz = wxStaticBoxSizer:new(?wxVERTICAL, Dialog, [{label, J("Address Hash")}]),
    AddressTx = wxTextCtrl:new(Dialog, ?wxID_ANY),
    _ = wxSizer:add(AddressSz, AddressTx, zxw:flags(wide)),
    ButtSz = wxBoxSizer:new(?wxHORIZONTAL),
    Affirm = wxButton:new(Dialog, ?wxID_OK),
    Cancel = wxButton:new(Dialog, ?wxID_CANCEL),
    _ = wxBoxSizer:add(ButtSz, Affirm, zxw:flags(wide)),
    _ = wxBoxSizer:add(ButtSz, Cancel, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, AddressSz, zxw:flags(wide)),
    _ = wxSizer:add(Sizer, ButtSz, zxw:flags(wide)),
    ok = wxDialog:setSizer(Dialog, Sizer),
    ok = wxBoxSizer:layout(Sizer),
    ok = wxDialog:setSize(Dialog, {500, 200}),
    ok = wxDialog:center(Dialog),
    ok = wxStyledTextCtrl:setFocus(AddressTx),
    Choice =
        case wxDialog:showModal(Dialog) of
            ?wxID_OK ->
                case wxTextCtrl:getValue(AddressTx) of
                    "" -> cancel;
                    A  -> {ok, A}
                end;
            ?wxID_CANCEL ->
                cancel
        end,
    ok = wxDialog:destroy(Dialog),
    case Choice of
        {ok, Address} -> load2(State, Address);
        cancel        -> State
    end.

load2(State, Address) ->
    case hz:contract_source(Address) of
        {ok, Source} ->
            load3(State, Address, Source);
        Error ->
            ok = handle_troubling(Error, State),
            State
    end.

load3(State, Address, Source) ->
    % TODO: Compile on load and verify the deployed hash for validity.
    case aeso_compiler:from_string(Source, [{aci, json}]) of
        {ok, Build = #{aci := ACI}} ->
            {Defs = #{functions := Funs}, ConIfaces} = find_main(ACI),
            Callable = lom:delete(name, <<"init">>, Funs),
            FunDefs = {maps:put(functions, Callable, Defs), ConIfaces},
            ok = tell(info, "Compilation Succeeded!~n~tp~n~n~tp", [Build, FunDefs]),
            add_instance_page(State, Address, Source);
        Other ->
            ok = tell(info, "Compilation Failed!~n~tp", [Other]),
            State
    end.

add_instance_page(State = #s{cons = {Consbook, Pages}, j = J}, Address, Source) ->
    Window = wxWindow:new(Consbook, ?wxID_ANY),
    PageSz = wxBoxSizer:new(?wxVERTICAL),

    ProgSz = wxBoxSizer:new(?wxHORIZONTAL),

    CodeSz = wxStaticBoxSizer:new(?wxVERTICAL, Window, [{label, J("Contract Source")}]),
    CodeTxStyle = {style,     ?wxTE_MULTILINE
                          bor ?wxTE_PROCESS_TAB
                          bor ?wxTE_DONTWRAP
                          bor ?wxTE_READONLY},
    CodeTx = wxTextCtrl:new(Window, ?wxID_ANY, [CodeTxStyle]),
    Mono = wxFont:new(10, ?wxMODERN, ?wxNORMAL, ?wxNORMAL, [{face, "Monospace"}]),
    TextAt = wxTextAttr:new(),
    ok = wxTextAttr:setFont(TextAt, Mono),
    true = wxTextCtrl:setDefaultStyle(CodeTx, TextAt),
    ok = wxTextCtrl:setValue(CodeTx, Source),
    _ = wxSizer:add(CodeSz, CodeTx, zxw:flags(wide)),

    ScrollWin = wxScrolledWindow:new(Window),
    FunSz = wxStaticBoxSizer:new(?wxVERTICAL, ScrollWin, [{label, J("Function Interfaces")}]),
    ok = wxWindow:setSizer(ScrollWin, FunSz),

    ConsSz = wxStaticBoxSizer:new(?wxVERTICAL, Window, [{label, J("Console")}]),
    ConsTxStyle = {style, ?wxTE_MULTILINE bor ?wxTE_READONLY},
    ConsTx = wxTextCtrl:new(Window, ?wxID_ANY, [ConsTxStyle]),
    _ = wxSizer:add(ConsSz, ConsTx, zxw:flags(wide)),

    _ = wxSizer:add(ProgSz, CodeSz,    [{proportion, 3}, {flag, ?wxEXPAND}]),
    _ = wxSizer:add(ProgSz, ScrollWin, [{proportion, 1}, {flag, ?wxEXPAND}]),

    _ = wxSizer:add(PageSz, ProgSz, [{proportion, 3}, {flag, ?wxEXPAND}]),
    _ = wxSizer:add(PageSz, ConsSz, [{proportion, 1}, {flag, ?wxEXPAND}]),

    ok = wxWindow:setSizer(Window, PageSz),
    ok = wxSizer:layout(PageSz),
    true = wxNotebook:addPage(Consbook, Window, Address, [{bSelect, true}]),
    Page = #c{id   = Address, win  = Window,
              code = CodeTx,  cons = ConsTx,
              funs = {ScrollWin, []}},
    NewPages = Pages ++ [Page],
    State#s{cons = {Consbook, NewPages}}.


edit(State) ->
    ok = tell(info, "EDIT clicked"),
    State.


close_instance(State) ->
    ok = tell(info, "CLOSE_INSTANCE clicked"),
    State.



%% (Somewhat silly) Data operations

store_nth(1, E, [_ | T]) -> [E | T];
store_nth(N, E, [H | T]) -> [H | store_nth(N - 1, T, E)].


drop_nth(1, [_ | T]) -> T;
drop_nth(N, [H | T]) -> [H | drop_nth(N - 1, T)].


keyfind_index(K, E, L) ->
    keyfind_index(K, E, 1, L).

keyfind_index(K, E, I, [H | T]) ->
    case element(E, H) =:= K of
        false -> keyfind_index(K, E, I + 1, T);
        true  -> {ok, I}
    end;
keyfind_index(_, _, _, []) ->
    error.
